package com.tzld.piaoquan.recommend.server.implement;


import com.alibaba.fastjson.JSONObject;
import com.ctrip.framework.apollo.spring.annotation.ApolloJsonValue;
import com.google.common.reflect.TypeToken;
import com.tzld.piaoquan.recommend.server.common.base.RankItem;
import com.tzld.piaoquan.recommend.server.framework.candidiate.Candidate;
import com.tzld.piaoquan.recommend.server.framework.common.User;
import com.tzld.piaoquan.recommend.server.framework.merger.MergeUtils;
import com.tzld.piaoquan.recommend.server.framework.merger.StrategyQueue;
import com.tzld.piaoquan.recommend.server.framework.recaller.BaseRecaller;
import com.tzld.piaoquan.recommend.server.framework.recaller.provider.RedisBackedQueue;
import com.tzld.piaoquan.recommend.server.framework.score.ScorerUtils;
import com.tzld.piaoquan.recommend.server.framework.utils.RedisSmartClient;
import com.tzld.piaoquan.recommend.server.gen.recommend.RecommendRequest;
import com.tzld.piaoquan.recommend.server.model.Video;
import com.tzld.piaoquan.recommend.server.service.rank.extractor.ExtractorUtils;
import com.tzld.piaoquan.recommend.server.service.rank.extractor.RankExtractorItemFeature;
import com.tzld.piaoquan.recommend.server.service.rank.extractor.RankExtractorUserFeature;
import com.tzld.piaoquan.recommend.server.util.CommonCollectionUtils;
import com.tzld.piaoquan.recommend.server.util.JSONUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.stereotype.Service;

import javax.annotation.Resource;
import java.text.SimpleDateFormat;
import java.util.*;
import java.util.stream.Collectors;

@Service
public class TopRecommendPipeline {


    @ApolloJsonValue("${rank.score.merge.weight:}")
    private Map<String, Double> mergeWeight;

    private static final Logger log = LoggerFactory.getLogger(TopRecommendPipeline.class);

    public static final String FILTER_CONF = "filter_config.conf";
    public static final String MERGE_CONF = "merge_config.conf";

    public static final String PREFIX = "";


    @Resource
    private RedisSmartClient client;
    @Resource
    public RedisTemplate<String, String> redisTemplate;

    public List<Video> feeds(final RecommendRequest requestData,
                             final int requestIndex,
                             final User userInfo) {
        // Step 1: Attention extraction
        List<RankItem> rankItems = feedByRec(requestData, requestIndex, userInfo);
        if (rankItems == null || rankItems.isEmpty()) {
            return new ArrayList<>();
        }
        List<String> rtFeaPartKey = new ArrayList<>(Arrays.asList("item_rt_fea_1day_partition", "item_rt_fea_1h_partition"));
        List<String> rtFeaPartKeyResult = this.redisTemplate.opsForValue().multiGet(rtFeaPartKey);
        Calendar calendar = Calendar.getInstance();
        String date = new SimpleDateFormat("yyyyMMdd").format(calendar.getTime());
        String hour = new SimpleDateFormat("HH").format(calendar.getTime());
        String rtFeaPart1h = date + hour;
        if (rtFeaPartKeyResult != null) {
            if (rtFeaPartKeyResult.get(1) != null) {
                rtFeaPart1h = rtFeaPartKeyResult.get(1);
            }
        }
        // 2 统计分
        String cur = rtFeaPart1h;
        List<String> datehours = new LinkedList<>();
        for (int i = 0; i < 24; ++i) {
            datehours.add(cur);
            cur = ExtractorUtils.subtractHours(cur, 1);
        }
        for (RankItem item : rankItems) {
            Map<String, String> itemBasicMap = item.getItemBasicFeature();
            Map<String, Map<String, Double>> itemRealMap = item.getItemRealTimeFeature();
            List<Double> views = getStaticData(itemRealMap, datehours, "view_pv_list_1h");
            List<Double> plays = getStaticData(itemRealMap, datehours, "play_pv_list_1h");
            List<Double> shares = getStaticData(itemRealMap, datehours, "share_pv_list_1h");
            List<Double> returns = getStaticData(itemRealMap, datehours, "p_return_uv_list_1h");
            List<Double> allreturns = getStaticData(itemRealMap, datehours, "return_uv_list_1h");

            List<Double> share2return = getRateData(returns, shares, 1.0, 1000.0);
            Double share2returnScore = calScoreWeight(share2return);
            List<Double> view2return = getRateData(returns, views, 1.0, 1000.0);
            Double view2returnScore = calScoreWeight(view2return);
            List<Double> view2play = getRateData(plays, views, 1.0, 1000.0);
            Double view2playScore = calScoreWeight(view2play);
            List<Double> play2share = getRateData(shares, plays, 1.0, 1000.0);
            Double play2shareScore = calScoreWeight(play2share);
            item.scoresMap.put("share2returnScore", share2returnScore);
            item.scoresMap.put("view2returnScore", view2returnScore);
            item.scoresMap.put("view2playScore", view2playScore);
            item.scoresMap.put("play2shareScore", play2shareScore);

            Double allreturnsScore = calScoreWeight(allreturns);
            item.scoresMap.put("allreturnsScore", allreturnsScore);

            // rov的趋势
            double trendScore = calTrendScore(view2return);
            item.scoresMap.put("trendScore", trendScore);

            // 新视频提取
            double newVideoScore = calNewVideoScore(itemBasicMap);
            item.scoresMap.put("newVideoScore", newVideoScore);

        }
        // 3 融合公式
        List<Video> result = new ArrayList<>();
        double alpha = this.mergeWeight.getOrDefault("alpha", 1.0);
        double beta = this.mergeWeight.getOrDefault("beta", 1.0);
        for (RankItem item : rankItems) {
            double trendScore = item.scoresMap.getOrDefault("trendScore", 0.0) > 0.0 ?
                    item.scoresMap.getOrDefault("trendScore", 0.0) : 0.0;
            double newVideoScore = item.scoresMap.getOrDefault("newVideoScore", 0.0) > 0.0 ?
                    item.scoresMap.getOrDefault("newVideoScore", 0.0) : 0.0;
            double score = item.getScoreStr() *
                    item.scoresMap.getOrDefault("share2returnScore", 0.0)
                    + alpha * trendScore
                    + beta * newVideoScore;
            Video video = new Video();
            video.setVideoId(Long.parseLong(item.getId()));
            video.setScore(score);
            video.setSortScore(score);
            video.setScoreStr(item.getScoreStr());
            video.setScoresMap(item.getScoresMap());
            result.add(video);
        }
        Collections.sort(result, Comparator.comparingDouble(o -> -o.getSortScore()));
        return result;
    }

    public List<Double> getStaticData(Map<String, Map<String, Double>> itemRealMap,
                                      List<String> datehours, String key){
        List<Double> views = new LinkedList<>();
        Map<String, Double> tmp = itemRealMap.getOrDefault(key, new HashMap<>());
        for (String dh : datehours){
            views.add(tmp.getOrDefault(dh, 0.0D) +
                    (views.isEmpty() ? 0.0: views.get(views.size()-1))
            );
        }
        return views;
    }

    public double calNewVideoScore(Map<String, String> itemBasicMap){
        double existenceDays = Double.valueOf(itemBasicMap.getOrDefault("existence_days", "30"));
        if (existenceDays > 8){
            return 0.0;
        }
        double score = 1.0 / (existenceDays + 5.0);
        return score;
    }
    public double calTrendScore(List<Double> data){
        double sum = 0.0;
        int size = data.size();
        for (int i=0; i<size-4; ++i){
            sum += data.get(i) - data.get(i+4);
        }
        if (sum * 10 > 0.6){
            sum = 0.6;
        }else{
            sum = sum * 10;
        }
        if (sum > 0){
            // 为了打断点
            sum = sum;
        }
        return sum;
    }

    public Double calScoreWeight(List<Double> data){
        Double up = 0.0;
        Double down = 0.0;
        for (int i=0; i<data.size(); ++i){
            up += 1.0 / (i + 1) * data.get(i);
            down += 1.0 / (i + 1);
        }
        return down > 1E-8? up / down: 0.0;
    }
    public List<Double> getRateData(List<Double> ups, List<Double> downs, Double up, Double down){
        List<Double> data = new LinkedList<>();
        for(int i=0; i<ups.size(); ++i){
            data.add(
                    (ups.get(i) + up) / (downs.get(i) + down)
            );
        }
        return data;
    }

    public List<RankItem> feedByRec(final RecommendRequest requestData,
                                    final int requestIndex,
                                    final User userInfo) {
        int recallNum = 200;

        // Step 1: Attention extraction
//        long timestamp = System.currentTimeMillis();
//        UserAttentionExtractorPipeline attentionExtractorPipeline = UserAttentionExtractorUtils.getAtttentionPipeline(UserAttentionExtractorUtils.BASE_CONF);
//        attentionExtractorPipeline.extractAttention(requestData, userInfo);

        // Step 2: create top queue
        StrategyQueue topQueue = MergeUtils.createTopQueue(MERGE_CONF, "top-queue");


        // Step 3: Candidate
        Map<String, Candidate> candidates = new HashMap<String, Candidate>();
        topQueue.candidate(candidates, recallNum, userInfo, requestData, 0, 0);


        // Step 4: Recalling & Basic Scoring
        RedisBackedQueue queueProvider = new RedisBackedQueue(client, 1000L);

        BaseRecaller recaller = new BaseRecaller(queueProvider);
        List<RankItem> items = recaller.recalling(requestData, userInfo, requestIndex, new ArrayList<>(candidates.values()));


        // Step 4: Advance Scoring
//        timestamp = System.currentTimeMillis();
//        ScorerPipeline scorerPipeline = getScorerPipeline(requestData);
//        items = scorerPipeline.scoring(requestData, userInfo, requestIndex, items);


        // Step 5: Merger
        MergeUtils.distributeItemsToMultiQueues(topQueue, items);
        topQueue.merge(recallNum * 3, userInfo, requestData, requestIndex, 0);

        // 多样性融合
//        List<RankItem> mergeItems = topQueue.getItems();
//        MergeUtils.diversityRerank(mergeItems, SimilarityUtils.getIsSameUserTagOrCategoryFunc(), recallNum, 6, 2);

        // Step 6: Global Rank & subList
        // TODO 前置和后置处理逻辑 hardcode，后续优化
        Map<String, String> sceneFeatureMap = getSceneFeature(requestData);
        Map<String, String> userFeatureMap = getUserFeatureMap(requestData, items);
        List<RankItem> rovRecallRankNewallScore = ScorerUtils.getScorerPipeline(ScorerUtils.BASE_CONF_NEW_FEED)
                .scoring(sceneFeatureMap, userFeatureMap, items);

        return rovRecallRankNewallScore;
    }

    private Map<String, String> getUserFeatureMap(RecommendRequest param, List<RankItem> rankItems) {
        Map<String, String> userFeatureMap = new HashMap<>(64);
        if (param.getMid() != null && !param.getMid().isEmpty()){
            String midKey = "user_info_4video_" + param.getMid();
            String userFeatureStr = redisTemplate.opsForValue().get(midKey);
            if (userFeatureStr != null){
                try{
                    userFeatureMap = JSONUtils.fromJson(userFeatureStr,
                            new TypeToken<Map<String, String>>() {},
                            userFeatureMap);
                }catch (Exception e){
                    log.error(String.format("parse user json is wrong in {} with {}", this.getClass().getSimpleName(), e));
                }
            }else{
                JSONObject obj = new JSONObject();
                obj.put("name", "user_key_in_model_is_null");
                obj.put("class", this.getClass().getSimpleName());
                log.info(obj.toString());
//                return videos;
            }
        }
        final Set<String> userFeatureSet = new HashSet<>(Arrays.asList(
                "machineinfo_brand", "machineinfo_model", "machineinfo_platform", "machineinfo_system",
                "u_1day_exp_cnt", "u_1day_click_cnt", "u_1day_share_cnt", "u_1day_return_cnt",
                "u_3day_exp_cnt", "u_3day_click_cnt", "u_3day_share_cnt", "u_3day_return_cnt"
        ));
        Iterator<Map.Entry<String, String>> iterator = userFeatureMap.entrySet().iterator();
        while (iterator.hasNext()) {
            Map.Entry<String, String> entry = iterator.next();
            if (!userFeatureSet.contains(entry.getKey())) {
                iterator.remove();
            }
        }
        Map<String, String> f1 = RankExtractorUserFeature.getOriginFeature(userFeatureMap,
                new HashSet<String>(Arrays.asList(
                        "machineinfo_brand", "machineinfo_model", "machineinfo_platform", "machineinfo_system"
                ))
        );
        Map<String, String> f2 = RankExtractorUserFeature.getUserRateFeature(userFeatureMap);
        Map<String, String> f3 = RankExtractorUserFeature.cntFeatureChange(userFeatureMap,
                new HashSet<String>(Arrays.asList(
                        "u_1day_exp_cnt", "u_1day_click_cnt", "u_1day_share_cnt", "u_1day_return_cnt",
                        "u_3day_exp_cnt", "u_3day_click_cnt", "u_3day_share_cnt", "u_3day_return_cnt"
                ))
        );
        f1.putAll(f2);
        f1.putAll(f3);
        log.info("userFeature in model = {}", JSONUtils.toJson(f1));

        // 2-1: item特征处理
        final Set<String> itemFeatureSet = new HashSet<>(Arrays.asList(
                "total_time", "play_count_total",
                "i_1day_exp_cnt", "i_1day_click_cnt", "i_1day_share_cnt", "i_1day_return_cnt",
                "i_3day_exp_cnt", "i_3day_click_cnt", "i_3day_share_cnt", "i_3day_return_cnt"
        ));

        List<Long> videoIds = CommonCollectionUtils.toListDistinct(rankItems, RankItem::getVideoId);
        List<String> videoFeatureKeys = videoIds.stream().map(r-> "video_info_" + r)
                .collect(Collectors.toList());
        List<String> videoFeatures = redisTemplate.opsForValue().multiGet(videoFeatureKeys);
        if (videoFeatures != null){
            for (int i=0; i<videoFeatures.size(); ++i){
                String vF = videoFeatures.get(i);
                Map<String, String> vfMap = new HashMap<>();
                if (vF == null){
                    continue;
                }
                try{
                    vfMap = JSONUtils.fromJson(vF, new TypeToken<Map<String, String>>() {}, vfMap);
                    Map<String, String> vfMapCopy = new HashMap<>(vfMap);
                    rankItems.get(i).setItemBasicFeature(vfMapCopy);
                    Iterator<Map.Entry<String, String>> iteratorIn = vfMap.entrySet().iterator();
                    while (iteratorIn.hasNext()) {
                        Map.Entry<String, String> entry = iteratorIn.next();
                        if (!itemFeatureSet.contains(entry.getKey())) {
                            iteratorIn.remove();
                        }
                    }
                    Map<String, String> f4 = RankExtractorItemFeature.getItemRateFeature(vfMap);
                    Map<String, String> f5 = RankExtractorItemFeature.cntFeatureChange(vfMap,
                            new HashSet<String>(Arrays.asList(
                                    "total_time", "play_count_total",
                                    "i_1day_exp_cnt", "i_1day_click_cnt", "i_1day_share_cnt", "i_1day_return_cnt",
                                    "i_3day_exp_cnt", "i_3day_click_cnt", "i_3day_share_cnt", "i_3day_return_cnt"))
                    );
                    f4.putAll(f5);
                    rankItems.get(i).setFeatureMap(f4);
                }catch (Exception e){
                    log.error(String.format("parse video json is wrong in {} with {}", this.getClass().getSimpleName(), e));
                }
            }
        }
        // 2-2: item 实时特征处理
        List<String> rtFeaPartKey = new ArrayList<>(Arrays.asList("item_rt_fea_1day_partition", "item_rt_fea_1h_partition"));
        List<String> rtFeaPartKeyResult = this.redisTemplate.opsForValue().multiGet(rtFeaPartKey);
        Calendar calendar = Calendar.getInstance();
        String date = new SimpleDateFormat("yyyyMMdd").format(calendar.getTime());
        String hour = new SimpleDateFormat("HH").format(calendar.getTime());
        String rtFeaPart1day = date + hour;
        String rtFeaPart1h = date + hour;
        if (rtFeaPartKeyResult != null){
            if (rtFeaPartKeyResult.get(0) != null){
                rtFeaPart1day = rtFeaPartKeyResult.get(0);
            }
            if (rtFeaPartKeyResult.get(1) != null){
                rtFeaPart1h = rtFeaPartKeyResult.get(1);
            }
        }

        List<String> videoRtKeys1 = videoIds.stream().map(r-> "item_rt_fea_1day_" + r)
                .collect(Collectors.toList());
        List<String> videoRtKeys2 = videoIds.stream().map(r-> "item_rt_fea_1h_" + r)
                .collect(Collectors.toList());
        videoRtKeys1.addAll(videoRtKeys2);
        List<String> videoRtFeatures = this.redisTemplate.opsForValue().multiGet(videoRtKeys1);


        if (videoRtFeatures != null){
            int j = 0;
            for (RankItem item: rankItems){
                ++j;
                if (j >= rankItems.size()) {
                    continue;
                }
                Map<String, String> vfMap = new HashMap<>();
                Map<String, Map<String, Double>> vfMapNew = new HashMap<>();
                try {
                    String vF = videoRtFeatures.get(j);
                    if (vF == null) {
                        continue;
                    }
                    vfMap = JSONUtils.fromJson(vF, new TypeToken<Map<String, String>>() {
                    }, vfMap);
                    for (Map.Entry<String, String> entry : vfMap.entrySet()) {
                        String value = entry.getValue();
                        if (value == null) {
                            continue;
                        }
                        String[] var1 = value.split(",");
                        Map<String, Double> tmp = new HashMap<>();
                        for (String var2 : var1) {
                            String[] var3 = var2.split(":");
                            tmp.put(var3[0], Double.valueOf(var3[1]));
                        }
                        vfMapNew.put(entry.getKey(), tmp);
                    }
                }catch (Exception e){
                    log.error(String.format("parse video item_rt_fea_1day_ json is wrong in {} with {}",
                            this.getClass().getSimpleName(), e));
                }
                Map<String, String> f8 = RankExtractorItemFeature.getItemRealtimeRate(vfMapNew, rtFeaPart1day);
                item.getFeatureMap().putAll(f8);
            }
            for (RankItem item: rankItems){
                ++j;
                if (j >= rankItems.size()) {
                    continue;
                }
                Map<String, String> vfMap = new HashMap<>();
                Map<String, Map<String, Double>> vfMapNew = new HashMap<>();
                try {
                    String vF = videoRtFeatures.get(j);
                    if (vF == null) {
                        continue;
                    }
                    vfMap = JSONUtils.fromJson(vF, new TypeToken<Map<String, String>>() {
                    }, vfMap);

                    for (Map.Entry<String, String> entry : vfMap.entrySet()) {
                        String value = entry.getValue();
                        if (value == null) {
                            continue;
                        }
                        String[] var1 = value.split(",");
                        Map<String, Double> tmp = new HashMap<>();
                        for (String var2 : var1) {
                            String [] var3 = var2.split(":");
                            tmp.put(var3[0], Double.valueOf(var3[1]));
                        }
                        vfMapNew.put(entry.getKey(), tmp);
                    }
                    item.setItemRealTimeFeature(vfMapNew);
                }catch (Exception e){
                    log.error(String.format("parse video item_rt_fea_1h_ json is wrong in {} with {}",
                            this.getClass().getSimpleName(), e));
                }
                Map<String, String> f8 = RankExtractorItemFeature.getItemRealtimeRate(vfMapNew, rtFeaPart1h);
                item.getFeatureMap().putAll(f8);
            }
        }


        log.info("ItemFeature = {}", JSONUtils.toJson(videoFeatures));
        return userFeatureMap;
    }

    private Map<String, String> getSceneFeature(RecommendRequest param) {
        Map<String, String> sceneFeatureMap = new HashMap<>();
        String provinceCn = param.getProvince();
        provinceCn = provinceCn.replaceAll("省$", "");
        sceneFeatureMap.put("ctx_region", provinceCn);
        String city = param.getCity();
        if ("台北市".equals(city) |
                "高雄市".equals(city) |
                "台中市".equals(city) |
                "桃园市".equals(city) |
                "新北市".equals(city) |
                "台南市".equals(city) |
                "基隆市".equals(city) |
                "吉林市".equals(city) |
                "新竹市".equals(city) |
                "嘉义市".equals(city)
        ){
            ;
        }else{
            city = city.replaceAll("市$", "");
        }
        sceneFeatureMap.put("ctx_city", city);

        Calendar calendar = Calendar.getInstance();
        sceneFeatureMap.put("ctx_week", (calendar.get(Calendar.DAY_OF_WEEK) + 6) % 7 + "");
        sceneFeatureMap.put("ctx_hour", new SimpleDateFormat("HH").format(calendar.getTime()));

        return sceneFeatureMap;
    }




}
